module List-Ext =
struct
  open List

  function index <a : Type> (list : List <a>, n : Nat) : a =
    case list of
      Nil          => throw Index
    | Cons (x, xs) => if n == 0 then x else index <a> (xs, n - 1)
    end

  function all <a : Type> (prop : a -> Bool) : List <a> -> Bool =
    rec fun test (ns : List <a>) : Bool =>
      case ns of Nil           => true
              |  Cons (n, ns') => prop n && test ns'
              end

  function any <a : Type> (prop : a -> Bool) : List <a> -> Bool =
    rec fun test (ns : List <a>) : Bool =>
      case ns of Nil           => false
              |  Cons (n, ns') => prop n || test ns'
              end

  function between-step (l : Nat, l': Nat, u : Nat) : List <Nat> =
    if l #< l' then
      let val d = l' - l in
        (rec fun up (i : Nat) : List <Nat> =>
          if i ># u  then  Nil
                     else  Cons (i, up (i + d))) l
      end
    else
      let val d = l - l' in
        (rec fun down (i : Nat) : List <Nat> =>
          if i #< u  then  Nil
                     else  Cons (i, down (i - d))) l
      end

end // List-Ext
