// lazy evaluation

data State <a> = Empty | Transit | Value a

open Maybe

type Thunk <a> = () -> a

val force = value

data State <a> = Empty | Transit | Value a

exception Black-hole

function lazy <a> (comp : Thunk <a>) : Thunk <a> =
  let val box = ref Empty in
    { case !box of 
        Empty   => let val v = box := Transit; force comp in
                     box := Value v; v
                   end
      | Transit => throw Black-hole
      | Value v => v
      end }
  end

function lazy <a> (comp : Thunk <a>) : Thunk <a> =
  let val box = ref Nothing in
    { case !box of 
        Nothing => let val v = force comp in
                     box := Just v; v
                   end
      | Just v  => v
      end }
  end

// lazy (rec s => {force s + 1})
// rec s => lazy {force s + 1}

// Streams

data State <a> = Cons (a, Thunk <State <a>>)

type Stream <a> = Thunk <State <a>>

function head <a> (s : Stream <a>) : a =
  case force s of Cons (x, xs) => x end

function tail <a> (s : Stream <a>) : Stream <a> =
  case force s of Cons (x, xs) => xs end

function show-Stream <a> (s : Stream <a>) : () =
  let function until (i : Nat, s : Stream <a>) : () =
        if i >= 5000 then ()
        else
          System.IO.println (head s);
          until (i + 1, tail s)
  in until (0, s) end

function show-Stream <a> (n : Nat, s : Stream <a>) : () =
  if n == 0 then
    ()
  else 
    System.IO.println (head s);
    show-Stream <a> (n - 1, tail s)

val undef = lazy (rec s => { Cons (1, { force (tail s) } ) }) // Black-holing klappt noch nicht

function from (n : Nat) : Stream <Nat> =
  lazy { Cons (n, from (n + 1)) }

// carry = 0 \/ carry + 1

// recursively defined
function const (n : Nat) : Stream <Nat> =
//  lazy (rec fun s () : State <Nat> => Cons (n, s))
//  lazy (rec s => { Cons (n, s) })
//  rec s => lazy { Cons (n, s) } !!! Laufzeitfehler aber kein Typfehler
  rec s => lazy <State <Nat>> (fun () => Cons (n, s)) // core dumped: only functions or records of functions may be recursively defined

function merge <a> (s1 : Stream <a>, s2 : Stream <a>) : Stream <a> =
  lazy { Cons (head s1, merge (s2, tail s1)) }

function add (s1 : Stream <Nat>, s2 : Stream <Nat>) : Stream <Nat> =
  lazy { Cons (head s1 + head s2, add (tail s1, tail s2)) }

val carry =
//  lazy (rec s => merge (const 0, add (s, const 1)))
//  lazy (rec fun s () : State <Nat> => force (merge (const 0, add (s, const 1))))
//  lazy (rec s => { Cons (0, merge (add (s, const 1), const 0)) })
  lazy (rec s => { force (merge (const 0, add (s, const 1))) })

// show-Stream carry

// Primzahlen

function candidates (n : Nat) : Stream <Nat> =
  lazy { Cons (n, candidates (n + 2)) }

function filter <a> (p : a -> Bool, s : Stream <a>) : Stream <a> =
  lazy { if p (head s) then Cons (head s, filter <a> (p, tail s))
                       else force (filter <a> (p, tail s)) }
// !! der Test `p (head s)' darf nicht auerhalb der lazy {} Klammern stehen !!

function prime-wrt (s : Stream <Nat>) (n : Nat) : Bool =
  let val p = head s in
    p * p ># n || n % p /= 0 && prime-wrt (tail s) n
  end

val primes = lazy (rec ps => { Cons (2, filter (prime-wrt ps, candidates 3)) })
